from .LLMs import LLMModel

from enum import Enum
from PIL import Image
import base64
from io import BytesIO
from dataclasses import dataclass

from openai.types.chat import ChatCompletion


class Role(Enum):
    system = "system"
    user = "user"
    ai = "assistant"
    agent = "agent"
    tool = "tool"


@dataclass
class ToolCall:
    id: str
    name: str
    arguments: dict
    result: any = ""


class Chat_message:
    def __init__(self, 
                 text: str, 
                 images: list[Image.Image]=None, 
                 role: Role=Role.user, 
                 answerdBy: LLMModel=None, 
                 raw_response: ChatCompletion=None
                ) -> None:
        self.role = role
        self.text = text
        self.images = self._process_image(images=images)
        self.answeredBy: LLMModel = answerdBy
        
        self.tool_calls: list[ToolCall] = []
        self.raw_resoonse: ChatCompletion = raw_response


    def __str__(self) -> str:
        # ANSI color codes for blue, green, and reset (to default)
        BLUE = "\033[34m"
        GREEN = "\033[32m"
        RESET = "\033[0m"
        
        message = ""
        
        if self.role == Role.system:
            message = "---------------------- System ----------------------\n"
        elif self.role == Role.user:
            message = BLUE + "----------------------- User -----------------------\n" + RESET
        elif self.role == Role.ai:
            message = GREEN + "--------------------- Assistant --------------------\n" + RESET
        
        # Append text and reset color formatting at the end
        message += self.text + RESET + "\n"
        
        return message
        
    def _process_image(self, images: list):
        """
        Process a list of images by resizing them to maintain aspect ratio and then converting them to base64 format.
        
        Args:
            images (list): A list of image objects to be processed.

        Returns:
            list: A list of base64-encoded image strings if input is not None/empty, otherwise `None`.
        
        Note:
            - Images should be provided as a "list" even if there is only a single image to process.
        """
        if images == None:
            return None

        base64_images = []
        for image in images:
            if image.mode == "RGBA":
                image = image.convert("RGB")
            
            image = self._resize_image_aspect_ratio(image=image)
            image = self._convert_to_base64(image=image)
            base64_images.append(image)

        return base64_images
        
    def _convert_to_base64(self, image: Image) -> str:
        """
        Convert an image to a base64-encoded string.

        Args:
            image (Image): The image object to be converted to base64 format.

        Returns:
            str: The base64-encoded string representation of the image.

        Note:
            - The image format will default to 'JPEG' if the format is not specified.
        """
        buffered = BytesIO()
        format = image.format if image.format else 'JPEG'
        image.save(buffered, format=format)
        img_bytes = buffered.getvalue()
        img_base64 = base64.b64encode(img_bytes).decode('utf-8')
        
        return img_base64
    
    def _resize_image_aspect_ratio(self, image: Image, target_length=1024):
        """
        Resize an image to a target length while maintaining its aspect ratio.

        Args:
            image (Image): The image object to be resized.
            target_length (int, optional): The target length for the larger dimension (default is 1024).

        Returns:
            Image: The resized image object with maintained aspect ratio.

        Note:
            - The smaller dimension is scaled proportionally based on the larger dimension to maintain aspect ratio.
            - If the image's aspect ratio is non-square, the target_length is applied to the larger dimension.
        """
        
        width, height = image.size
        
        if width > height:
            new_width = target_length
            new_height = int((target_length / width) * height)
        else:
            new_height = target_length
            new_width = int((target_length / height) * width)

        resized_image = image.resize((new_width, new_height))
        
        return resized_image
    
        


from dataclasses import dataclass

@dataclass
class LLMModel:
    name: str
    input_cost: float = 0
    output_cost: float = 0


# OpenAI
gpt_4o = LLMModel(name='openai/gpt-4o', input_cost=2.5, output_cost=10.0)
gpt_4o_mini = LLMModel(name='openai/gpt-4o-mini', input_cost=0.15, output_cost=0.60)
gpt_4_1 = LLMModel(name='openai/gpt-4.1', input_cost=2, output_cost=8.0)
gpt_4_1_mini = LLMModel(name='openai/gpt-4.1-mini', input_cost=0.4, output_cost=1.6)
gpt_4_1_nano = LLMModel(name='openai/gpt-4.1-nano', input_cost=0.1, output_cost=0.4)
o4_mini = LLMModel(name='openai/o4-mini', input_cost=1.1, output_cost=4.4)
o4_mini_high = LLMModel(name='openai/o4-mini-high', input_cost=1.1, output_cost=4.4)
o3 = LLMModel(name='openai/o3', input_cost=10, output_cost=40)

# Anthropic
claude_3_7_sonnet = LLMModel(name='anthropic/claude-3.7-sonnet', input_cost=3.0, output_cost=15.0)
claude_3_7_sonnet_thinking = LLMModel(name='anthropic/claude-3.7-sonnet:thinking', input_cost=3.0, output_cost=15.0)
claude_3_5_haiku = LLMModel(name='anthropic/claude-3.5-haiku', input_cost=0.8, output_cost=4.0)

# Google
gemini_2_0_flash = LLMModel(name='google/gemini-2.0-flash-001', input_cost=0.1, output_cost=0.4)
gemini_2_0_flash_free = LLMModel(name='google/gemini-2.0-flash-exp:free', input_cost=0.1, output_cost=0.4)
gemini_2_5_flash = LLMModel(name='google/gemini-2.5-flash-preview', input_cost=0.15, output_cost=0.60)
gemini_2_5_flash_thinking = LLMModel(name='google/gemini-2.5-flash-preview:thinking', input_cost=0.15, output_cost=3.5)
gemini_2_5_pro = LLMModel(name='google/gemini-2.5-pro-preview-03-25', input_cost=1.25, output_cost=10)

# Deepseek
deepseek_v3_free = LLMModel(name='deepseek/deepseek-chat-v3-0324:free', input_cost=0, output_cost=0)
deepseek_v3 = LLMModel(name='deepseek/deepseek-chat-v3-0324', input_cost=0.3, output_cost=1.2)
deepseek_r1_free = LLMModel(name='deepseek/deepseek-r1:free', input_cost=0, output_cost=0)
deepseek_r1 = LLMModel(name='deepseek/deepseek-r1', input_cost=0.5, output_cost=2.2)

# xAI
grok_3_mini = LLMModel(name='x-ai/grok-3-mini-beta', input_cost=0.3, output_cost=0.5)
grok_3 = LLMModel(name='x-ai/grok-3-beta', input_cost=3, output_cost=15)

# Microsoft
mai_ds_r1_free = LLMModel(name="microsoft/mai-ds-r1:free", input_cost=0, output_cost=0)

# Others
llama_4_maverick_free = LLMModel(name="meta-llama/llama-4-maverick:free", input_cost=0, output_cost=0)
llama_4_scout = LLMModel(name="meta-llama/llama-4-scout", input_cost=0.11, output_cost=0.34)
mistral_small_3_1_24B_free = LLMModel(name="mistralai/mistral-small-3.1-24b-instruct:free", input_cost=0, output_cost=0)


from .Chat_message import *
from .OpenRouterProvider import *
from .LLMs import LLMModel

from dotenv import load_dotenv
import time
import json
from typing import Iterator, AsyncIterator
from pydantic import BaseModel


_base_system_prompt = """
It's [TIME] today.
You are an intelligent AI. You must follow the system_instruction below, which is provided by the user.

<system_instruction>
[SYSTEM_INSTRUCTION]
</system_instruction>
"""

class Chatbot_manager:
    def __init__(self, system_prompt:str="", tools:list[tool_model]=[]) -> None:
        load_dotenv()
        
        self._memory: list[Chat_message] = []
        self.tools: list[tool_model] = tools
        self.set_system_prompt(prompt=system_prompt)
        
    def set_system_prompt(self, prompt: str):
        m, d, y = time.localtime()[:3]

        system_prompt = _base_system_prompt
        system_prompt = system_prompt.replace("[TIME]", f"{m}/{d}/{y}")
        system_prompt = system_prompt.replace("[SYSTEM_INSTRUCTION]", prompt)
        
        self._system_prompt = Chat_message(text=system_prompt, role=Role.system)
        
    def clear_memory(self):
        self._memory = []
        
    def print_memory(self):
        print("\n--------------------- Chatbot memory ---------------------")
        print(f"system   : {self._system_prompt.text}")

        for message in self._memory:
            role = message.role.value
            text = message.text.strip()
            
            reset_code = "\033[0m"
            role_str = f"{role.ljust(9)}:"
            indent = " " * len(role_str)
            lines = text.splitlines()
            
            if role == "user":
                color_code = "\033[94m"  # blue
                if lines:
                    print(f"{color_code}{role_str}{reset_code} {lines[0]}")
                    for line in lines[1:]:
                        print(f"{color_code}{indent}{reset_code} {line}")
                else:
                    print(f"{color_code}{role_str}{reset_code}")
            
            elif role == "assistant":
                color_code = "\033[92m"  # green
                if lines:
                    print(f"{color_code}{role_str}{reset_code} {lines[0]}")
                    for line in lines[1:]:
                        print(f"{color_code}{indent}{reset_code} {line}")
                else:
                    print(f"{color_code}{role_str}{reset_code}")

            elif role == "tool":
                color_code = "\033[93m"  # orange
                print(f"{color_code}{role_str}{reset_code} ", end="")
                
                for tool in message.tool_calls:
                    print(f"{tool.name}({json.loads(tool.arguments)}), ", end="")
                print()
                
            else:
                color_code = "\033[0m"   # default color
                print("Print error: The role is invalid.")
            
        print("----------------------------------------------------------\n")

    def invoke(
        self, 
        model: LLMModel, 
        query: Chat_message, 
        tools: list[tool_model]=[], 
        provider:ProviderConfig=None,
        temperature: float=0.3
    ) -> Chat_message:
        self._memory.append(query)
        client = OpenRouterProvider()
        reply = client.invoke(
            model=model,
            temperature=temperature,
            system_prompt=self._system_prompt,
            querys=self._memory,
            tools=self.tools + tools,
            provider=provider,
        )
        reply.answeredBy = model
        self._memory.append(reply)

        if reply.tool_calls:
            for requested_tool in reply.tool_calls:
                args = requested_tool.arguments
                if isinstance(args, str):
                    args = json.loads(args)

                for tool in (self.tools + tools):
                    if tool.name == requested_tool.name:
                        result = tool(**args)
                        requested_tool.result = result
                        break
                else:
                    print("Tool Not found", requested_tool.name)
                    return reply
                    
            reply = client.invoke(
                model=model,
                system_prompt=self._system_prompt,
                querys=self._memory,
                tools=self.tools + tools,
                provider=provider
            )
            
            reply.answeredBy = model
            self._memory.append(reply)

        return reply
    
    def invoke_stream(
        self, 
        model: LLMModel, 
        query: Chat_message, 
        tools: list[tool_model]=[], 
        provider:ProviderConfig=None,
        temperature: float=0.3
    ) -> Iterator[str]:
        self._memory.append(query)
        client = OpenRouterProvider()
        generator = client.invoke_stream(
            model=model,
            temperature=temperature,
            system_prompt=self._system_prompt,
            querys=self._memory,
            tools=self.tools + tools,
            provider=provider
        )
        
        text = ""
        for token in generator:
            text += token.choices[0].delta.content
            yield token.choices[0].delta.content

        self._memory.append(Chat_message(text=text, role=Role.ai, answerdBy=LLMModel))
        
    async def async_invoke(
        self, 
        model: LLMModel, 
        query: Chat_message, 
        tools: list[tool_model] = [], 
        provider: ProviderConfig = None,
        temperature: float=0.3
    ) -> Chat_message:
        self._memory.append(query)
        client = OpenRouterProvider()
        reply = await client.async_invoke(
            model=model,
            temperature=temperature,
            system_prompt=self._system_prompt,
            querys=self._memory,
            tools=self.tools + tools,
            provider=provider
        )
        reply.answeredBy = model
        self._memory.append(reply)

        if reply.tool_calls:
            for requested_tool in reply.tool_calls:
                args = requested_tool.arguments
                if isinstance(args, str):
                    args = json.loads(args)

                for tool in (self.tools + tools):
                    if tool.name == requested_tool.name:
                        result = tool(**args)
                        requested_tool.result = result
                        break
                else:
                    print("Tool Not found", requested_tool.name)
                    return reply

            reply = await client.async_invoke(
                model=model,
                system_prompt=self._system_prompt,
                querys=self._memory,
                tools=self.tools + tools,
                provider=provider
            )
            reply.answeredBy = model
            self._memory.append(reply)

        return reply

    async def async_invoke_stream(
        self, 
        model: LLMModel, 
        query: Chat_message, 
        tools: list[tool_model] = [], 
        provider: ProviderConfig = None,
        temperature: float=0.3
    ) -> AsyncIterator[str]:
        self._memory.append(query)
        client = OpenRouterProvider()

        stream = client.async_invoke_stream(
            model=model,
            temperature=temperature,
            system_prompt=self._system_prompt,
            querys=self._memory,
            tools=self.tools + tools,
            provider=provider
        )

        text = ""
        async for chunk in stream:
            delta = chunk.choices[0].delta.content or ""
            text += delta
            yield delta

        self._memory.append(Chat_message(text=text, role=Role.ai, answerdBy=model))
        
    def structured_output(
        self, 
        model: LLMModel, 
        query: Chat_message, 
        provider:ProviderConfig=None, 
        json_schema: BaseModel=None,
        temperature: float=0.3
    ) -> BaseModel:
        self._memory.append(query)
        client = OpenRouterProvider()
        reply = client.structured_output(
            model=model,
            temperature=temperature,
            system_prompt=self._system_prompt,
            querys=self._memory,
            provider=provider,
            json_schema=json_schema
        )
        return reply



import inspect
from functools import wraps
from typing import get_type_hints, get_origin, get_args


class tool_model:
    def __init__(self, func):
        self.func = func
        self.name = func.__name__

        sig = inspect.signature(func)
        type_hints = get_type_hints(func)

        type_map = {
            int:   "integer",
            str:   "string",
            float: "number",
            bool:  "boolean",
        }

        properties = {}
        required = []

        # Build JSON schema properties
        for name, param in sig.parameters.items():
            anno = type_hints.get(name, None)
            origin = get_origin(anno)

            if origin is list:
                # Handle list element types
                (elem_type,) = get_args(anno)
                json_item_type = type_map.get(elem_type, "string")
                schema = {
                    "type": "array",
                    "items": {"type": json_item_type},
                    "description": name,
                }
            else:
                # Primitive types
                json_type = type_map.get(anno, "string")
                schema = {
                    "type": json_type,
                    "description": name,
                }

            properties[name] = schema
            if param.default is inspect._empty:
                required.append(name)

        # Attach tool definition metadata
        self.tool_definition = {
            "type": "function",
            "function": {
                "name": func.__name__,
                "description": func.__doc__,
                "parameters": {
                    "type": "object",
                    "properties": properties,
                    "required": required,
                    "additionalProperties": False,
                },
                "strict": True,
            },
        }

        wraps(func)(self)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)
    
    



# structured output
# https://note.com/brave_quince241/n/n60a5759c8f05

import logging
from .Chat_message import *
from .Tool import tool_model
from .LLMs import *

from openai import OpenAI, AsyncOpenAI
from openai.types.chat import ChatCompletionChunk
from dotenv import load_dotenv
import os, time
from dataclasses import dataclass, field, asdict
from typing import List, Optional, Literal, Iterator, AsyncIterator
from pprint import pprint
from pydantic import BaseModel


# エラーのみ表示、詳細なトレースバック付き
logging.basicConfig(level=logging.ERROR, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@dataclass
class ProviderConfig:
    order: Optional[List[str]] = None
    allow_fallbacks: bool = None
    require_parameters: bool = None
    data_collection: Literal["allow", "deny"] = None
    only: Optional[List[str]] = None
    ignore: Optional[List[str]] = None
    quantizations: Optional[List[str]] = None
    sort: Optional[Literal["price", "throughput"]] = None
    max_price: Optional[dict] = None

    def to_dict(self) -> dict:
        return {k: v for k, v in asdict(self).items() if v is not None}


class OpenRouterProvider:
    def __init__(self) -> None:
        load_dotenv()
        api_key = os.getenv("OPENROUTER_API_KEY")
        if not api_key:
            logger.error("OPENROUTER_API_KEY is not set in environment variables.")
        self.client = OpenAI(
            base_url="https://openrouter.ai/api/v1",
            api_key=api_key,
        )
        self.async_client = AsyncOpenAI(
            base_url="https://openrouter.ai/api/v1",
            api_key=api_key,
        )

    def make_prompt(
        self, 
        system_prompt: Chat_message,
        querys: list[Chat_message]
    ) -> list[dict]:
        messages = [{"role": "system", "content": system_prompt.text}]

        for query in querys:
            if query.role == Role.user:
                if query.images is None:
                    messages.append({"role": "user", "content": query.text})
                else:
                    content = [{"type": "text", "text": query.text}]
                    for img in query.images[:50]:
                        content.append(
                            {"type": "image_url",
                             "image_url": {"url": f"data:image/jpeg;base64,{img}"}})
                    messages.append({"role": "user", "content": content})

            elif query.role == Role.ai or query.role == Role.tool:
                assistant_msg = {"role": "assistant"}
                assistant_msg["content"] = query.text or None

                if query.tool_calls:
                    assistant_msg["tool_calls"] = [
                        {
                            "id": str(t.id),
                            "type": "function",
                            "function": {
                                "name": t.name,
                                "arguments": t.arguments
                            }
                        }
                        for t in query.tool_calls
                    ]
                messages.append(assistant_msg)

            for t in query.tool_calls:
                messages.append({
                    "role": "tool",
                    "tool_call_id": str(t.id),
                    "content": str(t.result)
                })
            
        return messages

    def invoke(
        self, 
        model: LLMModel, 
        system_prompt: Chat_message, 
        querys: list[Chat_message], 
        tools: list[tool_model] = [], 
        provider: ProviderConfig = None,
        temperature: float = 0.3
    ) -> Chat_message:
        try:
            messages = self.make_prompt(system_prompt, querys)

            tool_defs = [tool.tool_definition for tool in tools] if tools else None
            provider_dict = provider.to_dict() if provider else None
            
            response = self.client.chat.completions.create(
                model=model.name,
                temperature=temperature,
                messages=messages,
                tools=tool_defs,
                extra_body={"provider": provider_dict},
            )

            reply = Chat_message(text=response.choices[0].message.content, role=Role.ai, raw_response=response)

            if response.choices[0].message.tool_calls:
                reply.role = Role.tool
                for tool in response.choices[0].message.tool_calls:
                    reply.tool_calls.append(ToolCall(id=tool.id, name=tool.function.name, arguments=tool.function.arguments))
            return reply

        except Exception as e:
            logger.exception(f"An error occurred while invoking the model: {e.__class__.__name__}: {str(e)}")
            return Chat_message(text="Fail to get response. Please see the error message.", role=Role.ai, raw_response=None)
        
    def invoke_stream(
        self, 
        model: LLMModel, 
        system_prompt: Chat_message, 
        querys: list[Chat_message], 
        tools: list[tool_model] = [], 
        provider: ProviderConfig = None,
        temperature: float = 0.3
    ) -> Iterator[ChatCompletionChunk]:
        # chunk example
        # ChatCompletionChunk(id='gen-1746748260-mdKZLTs9QY7MmUxWKb8V', choices=[Choice(delta=ChoiceDelta(content='!', function_call=None, refusal=None, role='assistant', tool_calls=None), finish_reason=None, index=0, logprobs=None, native_finish_reason=None)], created=1746748260, model='openai/gpt-4o-mini', object='chat.completion.chunk', service_tier=None, system_fingerprint='fp_e2f22fdd96', usage=None, provider='OpenAI')
        
        # ChatCompletionChunk(id='gen-1746748260-mdKZLTs9QY7MmUxWKb8V', choices=[Choice(delta=ChoiceDelta(content='', function_call=None, refusal=None, role='assistant', tool_calls=None), finish_reason='stop', index=0, logprobs=None, native_finish_reason='stop')], created=1746748260, model='openai/gpt-4o-mini', object='chat.completion.chunk', service_tier=None, system_fingerprint='fp_e2f22fdd96', usage=None, provider='OpenAI')
        
        # ChatCompletionChunk(id='gen-1746748260-mdKZLTs9QY7MmUxWKb8V', choices=[Choice(delta=ChoiceDelta(content='', function_call=None, refusal=None, role='assistant', tool_calls=None), finish_reason=None, index=0, logprobs=None, native_finish_reason=None)], created=1746748260, model='openai/gpt-4o-mini', object='chat.completion.chunk', service_tier=None, system_fingerprint=None, usage=CompletionUsage(completion_tokens=54, prompt_tokens=61, total_tokens=115, completion_tokens_details=CompletionTokensDetails(reasoning_tokens=0), prompt_tokens_details={'cached_tokens': 0}), provider='OpenAI')
        
        try:
            messages = self.make_prompt(system_prompt, querys)

            tool_defs = [tool.tool_definition for tool in tools] if tools else None
            provider_dict = provider.to_dict() if provider else None

            response = self.client.chat.completions.create(
                model=model.name,
                temperature=temperature,
                messages=messages,
                tools=tool_defs,
                extra_body={"provider": provider_dict},
                stream=True
            )
            
            return response

        except Exception as e:
            logger.exception(f"An error occurred while invoking the model: {e.__class__.__name__}: {str(e)}")
            return Chat_message(text="Fail to get response. Please see the error message.", role=Role.ai, raw_response=None)

    async def async_invoke(
        self, model: LLMModel, 
        system_prompt: Chat_message, 
        querys: list[Chat_message], 
        tools: list[tool_model] = [], 
        provider: ProviderConfig = None,
        temperature: float = 0.3
    ) -> Chat_message:
        try:
            messages = self.make_prompt(system_prompt, querys)

            tool_defs = [tool.tool_definition for tool in tools] if tools else None
            provider_dict = provider.to_dict() if provider else None

            response = await self.async_client.chat.completions.create(
                model=model.name,
                temperature=temperature,
                messages=messages,
                tools=tool_defs,
                extra_body={"provider": provider_dict}
            )

            reply = Chat_message(text=response.choices[0].message.content, role=Role.ai, raw_response=response)

            if response.choices[0].message.tool_calls:
                reply.role = Role.tool
                for tool in response.choices[0].message.tool_calls:
                    reply.tool_calls.append(ToolCall(id=tool.id, name=tool.function.name, arguments=tool.function.arguments))
            return reply

        except Exception as e:
            logger.exception(f"An error occurred while asynchronously invoking the model: {e.__class__.__name__}: {str(e)}")
            return Chat_message(text="Fail to get response. Please see the error message.", role=Role.ai, raw_response=None)
        
    async def async_invoke_stream(
        self,
        model: LLMModel,
        system_prompt: Chat_message,
        querys: list[Chat_message],
        tools: list[tool_model] = [],
        provider: ProviderConfig = None,
        temperature: float = 0.3
    ) -> AsyncIterator[ChatCompletionChunk]:
        try:
            messages = self.make_prompt(system_prompt, querys)

            tool_defs = [tool.tool_definition for tool in tools] if tools else None
            provider_dict = provider.to_dict() if provider else None

            response = await self.async_client.chat.completions.create(
                model=model.name,
                temperature=temperature,
                messages=messages,
                tools=tool_defs,
                extra_body={"provider": provider_dict},
                stream=True
            )

            async for chunk in response:
                yield chunk

        except Exception as e:
            logger.exception(f"An error occurred while asynchronously streaming the model: {e.__class__.__name__}: {str(e)}")
            return
        
    def structured_output(
        self, 
        model: LLMModel, 
        system_prompt: Chat_message, 
        querys: list[Chat_message], 
        provider: ProviderConfig = None, 
        json_schema: BaseModel = None,
        temperature: float = 0.3
    ) -> BaseModel:
        try:
            messages = self.make_prompt(system_prompt, querys)
            provider_dict = provider.to_dict() if provider else None
            
            response = self.client.chat.completions.create(
                model=model.name,
                temperature=temperature,
                messages=messages,
                response_format={"type": "json_schema", "json_schema": {"name": json_schema.__name__, "schema": json_schema.model_json_schema()}},
                extra_body={"provider": provider_dict},
            )

            return json_schema.model_validate_json(response.choices[0].message.content)

        except Exception as e:
            logger.exception(f"An error occurred while invoking structured output: {e.__class__.__name__}: {str(e)}")
            return Chat_message(text="Fail to get response. Please see the error message.", role=Role.ai, raw_response=None)
        
    